DROP TABLE IF EXISTS mysql.user_account_history;
DROP TABLE IF EXISTS mysql.user_password_policy_violation;
DROP PROCEDURE IF EXISTS mysql.check_password_policy;
DROP PROCEDURE IF EXISTS mysql.check_password_expiration;
DROP PROCEDURE IF EXISTS mysql.check_password_reuse;
DROP PROCEDURE IF EXISTS mysql.record_password_history;
DROP EVENT IF EXISTS mysql.password_policy_check_event;


CREATE TABLE mysql.user_account_history (
 id BIGINT UNSIGNED PRIMARY KEY AUTO_INCREMENT,
 host CHAR(60) COLLATE utf8_bin NOT NULL DEFAULT '',
 user CHAR(16) COLLATE utf8_bin NOT NULL DEFAULT '',
 password CHAR(41) CHARACTER SET latin1 COLLATE latin1_bin NOT NULL DEFAULT '',
 created_date TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
 KEY (host, user, created_date DESC)
);

CREATE TABLE mysql.user_password_policy_violation (
 id BIGINT UNSIGNED PRIMARY KEY AUTO_INCREMENT,
 history_id BIGINT UNSIGNED NOT NULL,
 host CHAR(60) COLLATE utf8_bin NOT NULL DEFAULT '',
 user CHAR(16) COLLATE utf8_bin NOT NULL DEFAULT '',
 violation_type INT NOT NULL,
 violation_message VARCHAR(255),
 violation_reported_date TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
 KEY (history_id, violation_type)
);


DELIMITER &&
CREATE DEFINER = root@localhost
PROCEDURE mysql.check_password_policy()
 BEGIN
 DECLARE pwd_expiration_days INT DEFAULT 90;
 DECLARE pwd_reuse_threshold INT DEFAULT 10;
 DECLARE done, new_id INT DEFAULT FALSE;
 DECLARE l_user, l_host, l_password VARCHAR(100);
 DECLARE cur CURSOR FOR 
   SELECT user, host, password 
   FROM mysql.user
   WHERE plugin = ''
   OR plugin = 'mysql_native_password';
 DECLARE CONTINUE HANDLER FOR NOT FOUND SET done = TRUE;

 OPEN cur;

 read_loop: LOOP
  FETCH cur INTO l_user, l_host, l_password;
  IF done THEN
    LEAVE read_loop;
  END IF;
  SET new_id = FALSE;
  CALL mysql.record_password_history( l_user, l_host, l_password, new_id);
  IF new_id THEN
   IF l_password = '' THEN
    INSERT INTO mysql.user_password_policy_violation
     (history_id, user, host, violation_type, violation_message)
    VALUES
     (new_id, l_user, l_host, 1, CONCAT(l_user, '@', l_host, ' has no password.'));
   ELSE
    CALL mysql.check_password_reuse( new_id, l_user, l_host, l_password, pwd_reuse_threshold );
   END IF;
  ELSE
   CALL mysql.check_password_expiration( l_user, l_host, l_password, pwd_expiration_days );
  END IF;
 END LOOP;

 CLOSE cur;
END
&& 

CREATE DEFINER = root@localhost
PROCEDURE mysql.check_password_expiration(
   IN l_user VARCHAR(60), 
   IN l_host VARCHAR(60),
   IN l_pass VARCHAR(41),
   IN days_threshold INT)
 BEGIN
 DECLARE done INT DEFAULT FALSE;
 DECLARE l_password VARCHAR(41);
 DECLARE l_days INT;
 DECLARE new_id BIGINT UNSIGNED;
 DECLARE reminder_days_snooze INT DEFAULT 10;
 DECLARE cur CURSOR FOR 
   SELECT h.id, h.password, DATEDIFF(CURRENT_DATE(), DATE(h.created_date))
   FROM mysql.user_account_history h
   LEFT JOIN mysql.user_password_policy_violation v ON 
    (h.id = v.history_id AND v.violation_type = 2 
     AND v.violation_reported_date > NOW() - INTERVAL reminder_days_snooze DAY)
   WHERE h.user = l_user
   AND h.host = l_host
   AND v.id IS NULL
   ORDER BY h.created_date DESC LIMIT 1;
 DECLARE CONTINUE HANDLER FOR NOT FOUND SET done = TRUE;

 OPEN cur;

 read_loop: LOOP
  FETCH cur INTO new_id, l_password, l_days;
  IF done THEN
    LEAVE read_loop;
  END IF;
  IF l_password = l_pass AND l_days >= days_threshold THEN
   SET done = TRUE;
   INSERT INTO mysql.user_password_policy_violation
     (history_id, user, host, violation_type, violation_message)
   VALUES
     (new_id, l_user, l_host, 2, CONCAT(l_user, '@', l_host, ' has not changed password in ', l_days, ' days.'));
  END IF;
 END LOOP;

 CLOSE cur;
END
&& 

CREATE DEFINER = root@localhost
PROCEDURE mysql.check_password_reuse(
   IN new_id BIGINT UNSIGNED,
   IN l_user VARCHAR(60), 
   IN l_host VARCHAR(60),
   IN l_pass VARCHAR(41),
   IN reuse_check_threshold INT)
 BEGIN
 DECLARE done INT DEFAULT FALSE;
 DECLARE l_password VARCHAR(41);
 DECLARE l_date DATE;
 DECLARE first_record INT DEFAULT TRUE;
 DECLARE cur CURSOR FOR 
   SELECT password, DATE(created_date)
   FROM mysql.user_account_history
   WHERE user = l_user
   AND host = l_host
   ORDER BY created_date DESC;
 DECLARE CONTINUE HANDLER FOR NOT FOUND SET done = TRUE;

 OPEN cur;

 read_loop: LOOP
  FETCH cur INTO l_password, l_date;
  IF done THEN
    LEAVE read_loop;
  END IF;
  IF reuse_check_threshold > 0 THEN
   SET reuse_check_threshold = reuse_check_threshold - 1;
  END IF;
  IF (NOT first_record) AND l_password = l_pass THEN
   SET done = TRUE;
   INSERT INTO mysql.user_password_policy_violation
     (history_id, user, host, violation_type, violation_message)
   VALUES
     (new_id, l_user, l_host, 3, CONCAT(l_user, '@', l_host, ' reused password from ', l_date));
  END IF;
  SET first_record = FALSE;
  IF reuse_check_threshold = 0 THEN
   SET done = TRUE;
  END IF;
 END LOOP;

 CLOSE cur;
END
&& 

CREATE DEFINER = root@localhost
PROCEDURE mysql.record_password_history(
   IN l_user VARCHAR(60), 
   IN l_host VARCHAR(60),
   IN l_pass VARCHAR(41),
   INOUT new_id INT)
 BEGIN
 DECLARE done, found INT DEFAULT FALSE;
 DECLARE l_password VARCHAR(41);
 DECLARE cur CURSOR FOR 
   SELECT password 
   FROM mysql.user_account_history
   WHERE user = l_user
   AND host = l_host
   ORDER BY created_date DESC LIMIT 1;
 DECLARE CONTINUE HANDLER FOR NOT FOUND SET done = TRUE;

 OPEN cur;

 read_loop: LOOP
  FETCH cur INTO l_password;
  IF done THEN
    LEAVE read_loop;
  END IF;
  IF l_password = l_pass THEN
   SET found = TRUE;
  END IF;
 END LOOP;

 CLOSE cur;
 IF NOT found THEN
  INSERT INTO mysql.user_account_history
   (user, host, password)
  VALUES
   (l_user, l_host, l_pass);
  SET new_id = LAST_INSERT_ID();
 END IF;
END
&& 

DELIMITER ;


CREATE DEFINER = root@localhost
EVENT mysql.password_policy_check_event
ON SCHEDULE EVERY 1 DAY
DO 
 CALL mysql.check_password_policy();
 
 
 
 